/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#include "dns/dnscache.h"

#include <fstream>
#include <logfunc.hpp>
#include <filefunc.hxx>   // I2n::file_exists
#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/asio/placeholders.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>

using boost::bind;
using boost::posix_time::seconds;
using I2n::Logger::GlobalLogger;

namespace Config
{
    int SaveTimerSeconds = 60;
}



DnsCache::DnsCache(const IoServiceItem &io_serv,
                   const std::string &cache_file)
    : DataCache()
    , SaveTimer( *io_serv )
    , CacheFile( cache_file )
    , HasChanged( false )
{
    // load cache from file
    load_from_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( Config::SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsCache::schedule_save, this,
                                boost::asio::placeholders::error ) );
}


DnsCache::~DnsCache()
{
    // save one last time without re-scheduling the next save
    save_to_cachefile();

    // cancel save timer
    SaveTimer.cancel();
}


void DnsCache::schedule_save(const boost::system::error_code &error)
{
    // just in case: ensure SaveTimer is cancelled
    SaveTimer.cancel();  // (will do nothing if already expired/cancelled)

    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.error() << "DNS Cache: SaveTimer was cancelled "
                             << "--> no save and no re-schedule of saving!";
        return;
    }
    else if (error)
    {
        GlobalLogger.error() << "DNS Cache: Received error " << error
                             << " in schedule_save "
                             << "--> no save now but re-schedule saving";
    }
    else
        save_to_cachefile();

    // schedule next save
    (void) SaveTimer.expires_from_now( seconds( Config::SaveTimerSeconds ) );
    SaveTimer.async_wait( bind( &DnsCache::schedule_save, this,
                                boost::asio::placeholders::error ) );
}

void DnsCache::save_to_cachefile()
{
    if (!HasChanged)
    {
        GlobalLogger.info() << "DNS Cache: skip saving because has not changed";
        return;
    }
    else if (CacheFile.empty())
    {
        GlobalLogger.warning()
                           << "DNS Cache: skip saving because file name empty!";
        return;
    }

    try
    {
        std::ofstream ofs( CacheFile.c_str() );
        boost::archive::xml_oarchive oa(ofs);
        oa << boost::serialization::make_nvp("DataCache", DataCache);
        GlobalLogger.info() << "DNS Cache: saved to cache file " << CacheFile;

        HasChanged = false;
    }
    catch (std::exception &exc)
    {
        GlobalLogger.warning() << "Saving failed: " << exc.what();
    }
}


void DnsCache::load_from_cachefile()
{
    if (CacheFile.empty())
    {
        GlobalLogger.warning()
                  << "DNS Cache: cannot load because cache file name is empty!";
        return;
    }
    else if ( !I2n::file_exists(CacheFile) )
    {
        GlobalLogger.warning() << "DNS Cache: cannot load because cache file "
                               << CacheFile << " does not exist!";
        return;
    }
    try
    {
        HostAddressVec cache;

        std::ifstream ifs( CacheFile.c_str() );
        boost::archive::xml_iarchive ia(ifs);

        ia >> boost::serialization::make_nvp("DataCache", cache);
        GlobalLogger.info() << "DNS Cache: loaded from file " << CacheFile;
    }
    catch (boost::archive::archive_exception &exc)
    {
        GlobalLogger.warning() << "DNS Cache: archive exception loading from "
                               << CacheFile << ": " << exc.what();
    }
    catch (std::exception &exc)
    {
        GlobalLogger.warning() << "DNS Cache: exception while loading from "
                               << CacheFile << ": " << exc.what();
    }
}

/**
 * 
 * in case the cached set is equal to the given new one, the old one is kept
 *   with TTLs updated
 * @returns true if changed cache; returns false if new_data is same as cache
 */
void DnsCache::update(const std::string &hostname,
                      const HostAddressVec &new_data)
{
    GlobalLogger.info() << "DNS Cache: update IPs for " << hostname
                        << " to " << new_data.size() << "-list";
    DataCache[hostname] = new_data;
    HasChanged = true;
}


HostAddressVec& DnsCache::get_data(const std::string &hostname)
{
    GlobalLogger.info() << "DNS Cache: request IPs for " << hostname
                        << " --> " << DataCache[hostname].size() << "-list";
    return DataCache[hostname];
}

// (created using vim -- the world's best text editor)

