/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015

 with code copied from boost::net::dns::resolve.hpp
 by Andreas Haberstroh (andreas at ibusy dot com)
    from https://github.com/softwareace/Boost.DNS
 */

#include "dns/dnsresolver.h"

#include <stdint.h>

#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/net/dns.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <logfunc.hpp>

using I2n::Logger::GlobalLogger;
using boost::posix_time::seconds;
using boost::posix_time::minutes;

namespace Config
{
    const int ResolveTimeoutSeconds = 5;
    const int PauseBeforeRetrySeconds = 10;
    const int StaleDataLongtermMinutes = 15;
    const int DNS_PORT = 53;
    const int UniqueID = 0xaffe;
}

DnsResolver::DnsResolver(IoServiceItem &io_serv,
                         const std::string &hostname,
                         const DnsIpProtocol &protocol,
                         const DnsCacheItem cache,
                         const boost::asio::ip::address &name_server)
    : ResolverBase( io_serv, hostname, cache )
    , Socket( *io_serv )
    , ReceiveBuffer()
    , Protocol( protocol )
    , NameServer( name_server, Config::DNS_PORT )
    , ResolveTimeoutTimer( *io_serv )
    , PauseBeforeRetryTimer( *io_serv )
    , StaleDataLongtermTimer( *io_serv )
    , NextIpIndex( 0 )
    , RetryCount( 0 )
    , IsResolving( false )
{ }

DnsResolver::~DnsResolver()
{
    boost::system::error_code error;
    Socket.shutdown(boost::asio::ip::udp::socket::shutdown_both, error);  //both=send&receive
    if ( error )
        GlobalLogger.warning() << "Received error " << error
                               << " when shutting down socket for DNS";
    // in IcmpPinger always gave an error system:9
    // (probably EBADF: Bad file descriptor)

    Socket.close(error);
    if ( error )
        GlobalLogger.warning() << "Received error " << error
                               << " when closing socket for DNS";
}



//==============================================================================
// ASYNC RESOLVE
//==============================================================================

/**
 * copied here code from boost::net::dns::resolve.hpp, since want async
 * operation and that is used only internally, there
 */
void DnsResolver::do_resolve()
{
    // check if resolving already
    if (IsResolving)
    {
        GlobalLogger.info()
            << "Call to do_resolve ignored since resolving already";
        return;
    }

    // just to be sure: cancel timers
    ResolveTimeoutTimer.cancel();
    PauseBeforeRetryTimer.cancel();
    StaleDataLongtermTimer.cancel();

    // create DNS request
    boost::net::dns::message dns_message( ResolverBase::Hostname,
                                          boost::net::dns::type_all );
    dns_message.recursive(false);
    dns_message.action(boost::net::dns::message::query);
    dns_message.opcode(boost::net::dns::message::squery);
    dns_message.id(Config::UniqueID);
    boost::net::dns_buffer_t request_buffer;
    dns_message.encode(request_buffer);

    // setup receipt of reply
    Socket.async_receive_from(
            boost::asio::buffer(ReceiveBuffer.get_array()),
            NameServer,
            boost::bind( &DnsResolver::handle_dns_result, this,
                         boost::asio::placeholders::error,
                         boost::asio::placeholders::bytes_transferred)
    );

    // schedule timeout
    (void) ResolveTimeoutTimer.expires_from_now(
                                        seconds(Config::ResolveTimeoutSeconds));
    ResolveTimeoutTimer.async_wait( boost::bind(
                                      &DnsResolver::handle_resolve_timeout,
                                      this, boost::asio::placeholders::error) );

    // send dns request
    Socket.send_to( boost::asio::buffer(request_buffer.get_array()),
                    NameServer );
}


void DnsResolver::handle_dns_result(const boost::system::error_code &error,
                       const std::size_t bytes_transferred)
{
    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.info() << "DNS resolve operation was cancelled";
        bool was_success = false;
        finalize_resolve(was_success);
    }
    else if (error)
    {
        GlobalLogger.info() << "DNS resolve resulted in error " << error
                            << " --> treat like unavailable";
        handle_unavailable();
        return;
    }

    // next 3(+1) lines copied from boost/net/dns/resolver.hpp:
    // clamp the recvBuffer with the number of bytes transferred or decode buffr
    ReceiveBuffer.length(bytes_transferred);
    boost::net::dns::message result_message;
    result_message.decode( ReceiveBuffer );

    // work with a regular pointer to list of answers since result_message is
    //   owner of data and that exists until end of function
    // Items in answers list are shared_ptr to resource_base_t
    boost::net::dns::rr_list_t *answers = result_message.answers();
    if (answers->size() == 0)
        handle_unavailable();

    // loop over answers, remembering ips and cnames
    HostAddressVec result_ips;
    std::vector<std::string> result_cnames;
    using boost::net::dns::resource_base_t;
    BOOST_FOREACH( boost::shared_ptr<resource_base_t> rr_item, *answers )
    {
        GlobalLogger.debug() << std::showbase << std::hex
                             << static_cast<unsigned>(rr_item->rtype()) << ": ";
        uint32_t ttl = rr_item->ttl();
        boost::net::dns::type_t rr_type = rr_item->rtype();

        if (rr_type == boost::net::dns::type_a)
        {    // 'A' resource records carry IPv4 addresses
            if (Protocol == DNS_IPv6)
            {
                GlobalLogger.info() << "Ignoring IPv4 address because resolver "
                                    << "was configured to only use IPv6.";
                continue;
            }
            boost::asio::ip::address_v4 ip =
                ( dynamic_cast<boost::net::dns::a_resource *> (rr_item.get()) )
                ->address();
            result_ips.push_back( HostAddress(ip, ttl) );
        }
        else if (rr_type == boost::net::dns::type_a6)
        {   // 'AAAA' resource records carry IPv6 addresses
            if (Protocol == DNS_IPv4)
            {
                GlobalLogger.info() << "Ignoring IPv6 address because resolver "
                                    << "was configured to only use IPv4.";
                continue;
            }
            boost::asio::ip::address_v6 ip =
                ( dynamic_cast<boost::net::dns::a6_resource *> (rr_item.get()) )
                ->address();
            result_ips.push_back( HostAddress(ip, ttl) );
        }
        else if (rr_type == boost::net::dns::type_cname)
        {   // 'CNAME' resource records that carry aliases
            std::string cname =
                (dynamic_cast<boost::net::dns::cname_resource *>(rr_item.get()))
                ->canonicalname();
            result_cnames.push_back( cname );
        }
        else if (rr_type == boost::net::dns::type_ns)
            GlobalLogger.debug() << "NS resource";
        else if (rr_type == boost::net::dns::type_soa)
            GlobalLogger.debug() << "SOA resource";
        else if (rr_type == boost::net::dns::type_ptr)
            GlobalLogger.debug() << "ptr resource";
        else if (rr_type == boost::net::dns::type_hinfo)
            GlobalLogger.debug() << "hinfo resource";
        else if (rr_type == boost::net::dns::type_mx)
            GlobalLogger.debug() << "mx resource";
        else if (rr_type == boost::net::dns::type_txt)
            GlobalLogger.debug() << "txt resource";
        else if (rr_type == boost::net::dns::type_srv)
            GlobalLogger.debug() << "srv resource";
        else if (rr_type == boost::net::dns::type_axfr)
            GlobalLogger.debug() << "axfr resource";
        else
            GlobalLogger.debug() << "unknown resource type";
    }

    GlobalLogger.info() << "Have " << result_ips.size() << " IPs and "
                        << result_cnames.size() << " CNAMEs";

    // We expect either one single CNAME and no IPs or a list of IPs.
    // But deal with other cases as well
    if (result_ips.empty() && result_cnames.empty())
        handle_unavailable();   // we just got crap, this is a dead end
    else if ( !result_ips.empty() && !result_cnames.empty())
        GlobalLogger.warning() << "Have CNAMEs AND IPs --> deal with both!";

    BOOST_FOREACH( const std::string &cname, result_cnames )
        handle_cname(cname);  // will schedule another DNS call

    if ( !result_ips.empty() )
        handle_ips(result_ips);
}


void DnsResolver::handle_ips(const HostAddressVec &ips)
{
    // save in cache
    ResolverBase::update_cache( ips );

    // clean up
    bool was_success = true;
    finalize_resolve(was_success);
}


void DnsResolver::handle_unavailable()
{
    // schedule new attempt in quite a while
    StaleDataLongtermTimer.expires_from_now(
                                     minutes(Config::StaleDataLongtermMinutes));
    StaleDataLongtermTimer.async_wait(
            boost::bind( &DnsResolver::wait_timer_timeout_handler,
                         this, boost::asio::placeholders::error
            )
    );

    // for now, admit failure
    bool was_success = false;
    finalize_resolve(was_success);
}

void DnsResolver::handle_cname(const std::string &canonical_name)
{
    // get resolver for canonical name
    ResolverItem resolver = DnsMaster::get_instance()
                            ->get_resolver_for(canonical_name, Protocol);
    callback_type callback = boost::bind( &DnsResolver::cname_resolve_callback,
                                          this, canonical_name, _1, _2 );
    resolver->async_resolve( callback );

    stop_trying();
}


void DnsResolver::cname_resolve_callback(const std::string &canonical_name,
                                         const bool was_success,
                                         const int cname_count)
{
    if (was_success)
        // tell cache to return cname's ips if queried for our hostname
        ResolverBase::update_cache(
                             ResolverBase::get_cached_results(canonical_name) );
    else
        GlobalLogger.info() << "Cname resolution failed";

    finalize_resolve(was_success, cname_count+1);
}


void DnsResolver::finalize_resolve(const bool was_success,
                                   const int cname_count)
{
    // stop timers
    if (cname_count > 0)
        stop_trying();
    // else was called already from handle_cname

    // schedule callbacks, clearing callback list
    ResolverBase::schedule_callbacks(was_success, cname_count);

    // finalize
    GlobalLogger.notice() << "Done resolving"
                          << " with success = " << was_success
                          << " and cname_count = " << cname_count;
    IsResolving = false;
}

void DnsResolver::stop_trying()
{
    // cancel timers
    GlobalLogger.debug() << "Cancelling timers";
    ResolveTimeoutTimer.cancel();
    PauseBeforeRetryTimer.cancel();
    StaleDataLongtermTimer.cancel();

    // clean up
    RetryCount = 0;
}

void DnsResolver::handle_resolve_timeout(const boost::system::error_code &error)
{
    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
    {
        GlobalLogger.warning() << "Resolve timeout timer was cancelled!";
        return;
    }
    else if (error)
    {
        GlobalLogger.warning() << "resolve timeout handler received error "
                               << error;
        return;
    }

    GlobalLogger.notice() << "DNS resolving timed out";

    // increment timer
    ++RetryCount;

    if ( RetryCount > DnsMaster::get_instance()
                      ->get_max_address_resolution_attempts() )
    {
        handle_unavailable();
        RetryCount = 0;
    }
    else
    {   // schedule retry
        PauseBeforeRetryTimer.expires_from_now(
                seconds(Config::PauseBeforeRetrySeconds));
        PauseBeforeRetryTimer.async_wait(
                boost::bind( &DnsResolver::wait_timer_timeout_handler,
                             this, boost::asio::placeholders::error) );
    }
}

void DnsResolver::wait_timer_timeout_handler(
                                         const boost::system::error_code &error)
{
    if ( error ==  boost::asio::error::operation_aborted )   // cancelled
        GlobalLogger.warning() << "Resolve timeout timer was cancelled!";
    else if (error)
        GlobalLogger.warning() << "resolve timeout handler received error "
                               << error;
    else
    {
        GlobalLogger.info() << "Done waiting --> re-try resolve";
        do_resolve();
    }
}


//==============================================================================
// RETRIEVAL
//==============================================================================

HostAddress DnsResolver::get_next_ip()
{
    // get cached data
    HostAddressVec cached_data = ResolverBase::get_cached_results();

    // if no results cached, return default-constructed HostAddress (0.0.0.0)
    if ( cached_data.empty() )
    {
        HostAddress return_value;
        return return_value;
    }

    // check validity of index (cache may have changed since last call)
    if (NextIpIndex >= cached_data.size())
        NextIpIndex = 0;

    // return next IP
    return cached_data[NextIpIndex++];
}

bool DnsResolver::have_up_to_date_ip()
{
    // get cached data
    HostAddressVec cached_data = ResolverBase::get_cached_results();

    // get threshold
    uint32_t resolved_ip_ttl_threshold = static_cast<uint32_t>(
                   DnsMaster::get_instance()->get_resolved_ip_ttl_threshold() );

    // loop over addresses
    BOOST_FOREACH( const HostAddress &addr, cached_data )
    {
        uint32_t ttl = addr.get_ttl().get_updated_value();
        if ( ttl > resolved_ip_ttl_threshold )
            return true;
    }

    // if not returned true by now, we have tried all IPs without success
    return false;
}

int DnsResolver::get_resolved_ip_count()
{
    return ResolverBase::get_cached_results().size();
}

// (created using vim -- the world's best text editor)

